/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * A way in which some virtual POSIX-like system is embedded in some
 * other system, one that might or might not be an actual POSIX system.
 *
 * See the interface header comments for details.
 */

#ifndef VRPSX_EMBEDDING_H
#define VRPSX_EMBEDDING_H 1

#include "vrpsx_system.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* A way in which some virtual POSIX-like system is embedded in some
 * other system, one that might or might not be an actual POSIX system.
 * For example, it may be sandboxed, emulated, etc.  This interface
 * provides the relevant path translations, etc.
 *
 * Please note that even in situations where this interface extension
 * is provided, there is a good chance you may get an enosys error
 * because a particular operation is not meaningful/implemented.
 *
 * You can cast a pointer to this interface to (vrpsx_counted_identified_t *)
 * because this interface extends vrpsx_counted_identified_t.
 *
 * Extends: vrpsx_counted_identified_t.
 * Extends at nonzero offset: vrpsx_system_t.
 * Interface Lifetime: reference counted.
 * Thread Safety: thread-safe.
 */
typedef struct vrpsx_embedding_struct {
#define VRPSX_EMBEDDING_MEMBERS(VTBL)                                   \
    VRPSX_COUNTED_IDENTIFIED_MEMBERS(VTBL)                              \
                                                                        \
    /* An interface to the virtual POSIX-like system that is embedded;  \
     * vrpsx_embedding_t extends vrpsx_system_t through this member.    \
     *                                                                  \
     * NOTE: We extend vrpsx_system_t at a nonzero offset so that       \
     * vrpsx_embedding_vtbl_t does not have to include operations       \
     * from vrpsx_system_vtbl_t.  We wish to avoid such inclusion       \
     * so that backward-compatible expansion of vrpsx_system_vtbl_t     \
     * to include new members does not affect the offsets of the        \
     * members of vrpsx_embedding_vtbl_t, and thus break backward       \
     * compatibility of vrpsx_embedding_vtbl_t.                         \
     *                                                                  \
     * If at some point it is found that such compatibility concerns    \
     * are unimportant, then vrpsx_embedding could extend vrpsx_system  \
     * at a zero offset, which would eliminate any need to separately   \
     * extend the vrpsx_counted_identified interface.  (You might then  \
     * wish to adjust vrpsx_embedding_main to skip the extra pointer.)  \
     */                                                                 \
    vrpsx_system_t system;

    VRPSX_EMBEDDING_MEMBERS(struct vrpsx_embedding_vtbl_struct)
} vrpsx_embedding_t;

typedef struct vrpsx_embedding_end_struct {
} vrpsx_embedding_end_t;

typedef struct vrpsx_embedding_vtbl_struct {
#define VRPSX_EMBEDDING_VTBL_MEMBERS(TYPE)                              \
    VRPSX_COUNTED_IDENTIFIED_VTBL_MEMBERS(TYPE)                         \
                                                                        \
    /* Converts to an internal PID the external process identifier      \
     * that starts at outer_pid and has size outer_size, which must     \
     * be appropriately aligned for its type.                           \
     *                                                                  \
     * NOTE: Some external processes may have no internal analog        \
     * at all.  In such cases an esrch error will result.               \
     *                                                                  \
     * The type of external process identifier expected by              \
     * this operation is defined by the outer system, not the           \
     * implementation of this interface per se.  For example,           \
     * an embedding into another POSIX-like system as such must         \
     * assign type pid_t to the external process identifier.            \
     *                                                                  \
     * On error the new value of *inner_pid is unspecified.             \
     */                                                                 \
    vrpsx_errno_t (*internalize_pid)(                                   \
        TYPE *embedding,                                                \
        void const *outer_pid,                                          \
        size_t outer_size,                                              \
        vrpsx_pid_t *inner_pid);                                        \
                                                                        \
    /* Converts the given internal PID                                  \
     * to an external process identifier.                               \
     *                                                                  \
     * The type of external process identifier produced by              \
     * this operation is defined by the outer system, not the           \
     * implementation of this interface per se.  For example,           \
     * an embedding into another POSIX-like system as such must         \
     * assign type pid_t to the external process identifier.            \
     *                                                                  \
     * That external process identifier type may have alignment         \
     * requirements, in which case the allocator supplied by the        \
     * caller must satisfy those requirements (often by returning       \
     * a pointer to a preallocated variable).                           \
     */                                                                 \
    vrpsx_errno_t (*externalize_pid)(                                   \
        TYPE *embedding,                                                \
        vrpsx_pid_t inner_pid,                                          \
        vrpsx_allocator_t *allocator, /* singular allocation of *outer_pid; \
                                       * NULL means                     \
                                       * &embedding->system.allocator */ \
        void **outer_pid,    /* set to result, or to NULL on error */   \
        size_t *outer_size); /* set to size of result, or to 0 on error */ \
                                                                        \
    /* Converts an external pathname into an embedded POSIX pathname.   \
     * Both the input and output pathnames are NUL-terminated.          \
     *                                                                  \
     * The kind of external pathname expected by this operation is      \
     * defined by the outer system, not the implementation of this      \
     * interface per se.  For example, if the outer system is a POSIX   \
     * system, then the external pathname should be a POSIX pathname.   \
     *                                                                  \
     * If the outer system mandates ISO/IEC 10646, or a character set   \
     * that safely converts to UCS, then you must encode the external   \
     * pathname using UTF-8 before passing it to this operation.  Note  \
     * that this requirement does not apply to typical Linux systems,   \
     * where UTF-8 pathnames may be a good idea but the kernel actually \
     * accepts byte strings with few restrictions.  In that context the \
     * external pathname is just the pathname understood by the kernel. \
     *                                                                  \
     * Where practical this operation should keep relative paths        \
     * relative; however, that usually makes sense only if the outer    \
     * and inner working directories are kept synchronized, and in      \
     * any case beware that changes to the current working directory    \
     * may expose you to thread-safety problems.  We choose to keep     \
     * relative paths relative because converting from relative to      \
     * absolute is usually easier than the opposite conversion.         \
     *                                                                  \
     * Special Case: If the implementation notices that the inner path  \
     * will be the same as the outer path (byte for byte), then instead \
     * of allocating a copy of the input it may instead set *inner_path \
     * to NULL and return 0.  This outcome NOT an error.                \
     *                                                                  \
     * This operation is not required to validate its input             \
     * (though it is free to do so) and therefore a non-sandboxed       \
     * native implementation might unconditionally set *inner_path      \
     * to NULL and return 0.                                            \
     *                                                                  \
     * NOTE: Some external paths may have no internal analog            \
     * at all.  In such cases an enoent error will result.              \
     */                                                                 \
    vrpsx_errno_t (*internalize_path)(                                  \
        TYPE *embedding,                                                \
        char const *outer_path,                                         \
        vrpsx_allocator_t *allocator, /* singular allocation of *inner_path; \
                                       * NUL-terminated allocation;     \
                                       * NULL means                     \
                                       * &embedding->system.allocator */ \
        char **inner_path);  /* set to result, or to NULL on error      \
                              * or if output would equal input */       \
                                                                        \
    /* Converts to an embedded POSIX pathname to an external pathname.  \
     * Both the input and output pathnames are be NUL-terminated.       \
     *                                                                  \
     * The external pathname produced by this operation satisfies the   \
     * requirements described by "internalize_path" for its input.      \
     * See the documentation of that operation for details.             \
     *                                                                  \
     * Where practical this operation should keep relative paths        \
     * relative; see "internalize_path" for details.                    \
     *                                                                  \
     * Special Case: If the implementation notices that the outer path  \
     * will be the same as the inner path (byte for byte), then instead \
     * of allocating a copy of the input it may instead set *outer_path \
     * to NULL and return 0.  This outcome NOT an error.                \
     *                                                                  \
     * This operation is not required to validate its input             \
     * (though it is free to do so) and therefore a non-sandboxed       \
     * native implementation might unconditionally set *outer_path      \
     * to NULL and return 0.                                            \
     *                                                                  \
     * NOTE: There is no guarantee that this operation is the inverse   \
     * of internalize_path, though of course that is desirable, and if  \
     * not possible please try to make the round trip idempotent.       \
     */                                                                 \
    vrpsx_errno_t (*externalize_path)(                                  \
        TYPE *embedding,                                                \
        char const *inner_path,                                         \
        vrpsx_allocator_t *allocator, /* singular allocation of *outer_path; \
                                       * NUL-terminated allocation;     \
                                       * NULL means                     \
                                       * &embedding->system.allocator */ \
        char **outer_path);  /* set to result, or to NULL on error      \
                              * or if output would equal input */       \
                                                                        \
    /* Like internalize_path, but transforms an execution path          \
     * such as you might find in the PATH environment variable          \
     * of most operating systems.                                       \
     */                                                                 \
    vrpsx_errno_t (*internalize_execution_path)(                        \
        TYPE *embedding,                                                \
        char const *outer_path,                                         \
        vrpsx_allocator_t *allocator,                                   \
        char **inner_path);                                             \
                                                                        \
    /* Like externalize_path, but transforms an execution path          \
     * such as you might find in the PATH environment variable          \
     * of most operating systems.                                       \
     */                                                                 \
    vrpsx_errno_t (*externalize_execution_path)(                        \
        TYPE *embedding,                                                \
        char const *inner_path,                                         \
        vrpsx_allocator_t *allocator,                                   \
        char **outer_path);                                             \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    vrpsx_embedding_end_t const *embedding_end;

    VRPSX_EMBEDDING_VTBL_MEMBERS(vrpsx_embedding_t)
} vrpsx_embedding_vtbl_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_EMBEDDING_H */
